#!/usr/bin/env python2.1
# -*- python -*-
#
#      Copyright (C) 2002 Robert Muth <robert@muth.org>
#
#      This program is free software; you can redistribute it and/or modify
#      it under the terms of the GNU General Public License as published by
#      the Free Software Foundation; version 2 of June 1991.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU General Public License for more details.
#
#      You should have received a copy of the GNU General Public License
#      along with this program in form of the file COPYING; 
#      if not, write to the 
#
#      Free Software Foundation, Inc. <http://www.fsf.org>
#      59 Temple Place, Suite 330, 
#      Boston, MA 02111-1307  USA
#
######################################################

__version__ = "Cjukebox $Revision: 1.15 $"

#######################################################
## CONFIG
#######################################################

ConfigSpecialPlaylistRoot = "/home/music/0_Playlist"
ConfigMusicRoot = "/home/music"
ConfigLogFile   = "cjukebox.log"
ConfigShowInfoScript = "./cjukebox.info.sh"

#######################################################
##  STANDARD IMPORTS
#######################################################

import os
import sys
import time
import getopt
import signal
import string
#import select
#import re
import curses
import curses.textpad
import whrandom
from curses import panel

#######################################################
## LOCAL IMPORTS
#######################################################


import musicus
import playlist
import song 
#import cddb

#######################################################
## 
#######################################################
logfile = None

def Message(prefix,message):
    global logfile
    if not logfile:
        logfile = open(ConfigLogFile,"wa")
    logfile.write(prefix + ": " + message)

def Debug(message):
    Message("D",message)

def Info(message):
    Message("I",message)

def Warning(message):
    Message("W",message)

def Error(message):
    Message("E",message)
    sys.exit(-1)
    
#######################################################
## 
#######################################################

class Keymap:
    def __init__(self):
        self.methods = [None] * curses.KEY_MAX

    def bind(self, key, method, args, desc, focus):
        self.methods[key] = (method, args, desc, focus)

    def is_bound(self,key):
        if self.methods[key] == None:
            return 0
        return 1

    def process(self, key, focus):
        if self.is_bound(key):
            method, args, desc, f  = self.methods[key]
            if focus != 0 or f == 0:
                apply(method, args)
                return 1
        return 0

    def help_string(self):
        message = []
        for k in range(curses.KEY_MAX):
            if self.methods[k] == None:
                continue
            method, args, desc, focus  = self.methods[k]
            if desc:
                if focus:
                    line = "* "
                else:
                    line = "  "
                line += desc
                message.append(line)
        return message

#######################################################
## 
#######################################################

class TimerEvents:
    def __init__(self):
        self.tag = 0
        self.events = {}

    def add_onetime(self, timeout, func, args):
        self.tag = self.tag + 1
        self.events[self.tag] = (func, args, time.time() + timeout, -1)
        return self.tag

    def add_periodic(self, period, func, args):
        self.tag = self.tag + 1
        self.events[self.tag] = (func, args, time.time() + period, period)
        return self.tag

    def remove(self, tag):
        if self.events.has_key(tag):
            del self.events[tag]

    def poll(self):
        now = time.time()

        for tag, (func, args, timeout, period) in self.events.items():
            if now >= timeout:
                apply(func, args)
                if period < 0.0:
                    self.remove(tag)
                else:
                    self.events[tag] = (func, args, now + period, period)

#######################################################
## 
#######################################################

GlobalLine = "=" * 60

class Window:

    ColorFocus = []
    ColorNoFocus = []

    global GlobalLine

    def __init__(self, parent, name, x, y, w, h):
        self._parent = parent
        self._children = []
        self._name = name
        self._keymap = Keymap()
        self._window = curses.newwin(h, w, y, x)
        self._color = Window.ColorNoFocus
        self._focus = 0

        if curses.has_colors():
            self._window.attron(Window.ColorNoFocus[0])
            
        if parent:
            parent.child_add(self)

            
    def child_add(self,child):
        self._children.append(child)
        
    def focus_on(self):
        self._focus = 1
        self._window.attron(curses.A_BOLD)
        self.redraw()
        
    def focus_off(self):
        self._focus = 0
        self._window.attroff(curses.A_BOLD)
        self.redraw()
        
    def has_focus(self):
        return self._focus
        
    def process_key(self, key):
        if self._keymap.process(key,self._focus):
            return 1
        for child in self._children:
            if child.process_key(key):
                return 1
        
    def help_strings(self,recursive=1,focusonly=0):
        message = []
        if focusonly and not self._focus:
            return message
 
        message.append(GlobalLine)
        message.append(self._name)
        message.append(GlobalLine)
        
        message += self._keymap.help_string()
        
        if recursive:
            for child in self._children:
                message += child.help_strings(recursive,focusonly)
        return message

#######################################################
## 
#######################################################

class ModalDialog:
    def __init__(self,x, y, w, h):
        self._window = curses.newwin(h, w, y, x)
        self._textbox = curses.textpad.Textbox(  self._window )
        self._textbox.edit()
        
#######################################################
## Generic Panel Class 
#######################################################

class Panel(Window):

    def __init__(self,parent,name, x, y, w, h):
        Window.__init__(self, parent, name, x, y, w, h)

        self._pan = panel.new_panel(self._window)
        self._screen_h = h - 2
        self._screen_w = w - 2
        return
    
    def draw_set_attributes(self,cursor=0):
        if cursor:
            self._window.attron(curses.A_REVERSE)
        else:
            self._window.attroff(curses.A_REVERSE)
        
#        if cursor and self._focus:
#            color = 1
#        elif cursor:
#            color = 2
#        else:
#            color = 0
        return 


    def draw_clear(self):
        self.draw_set_attributes()
        self._window.clear()

    def draw_header(self,pos,string):
        self.draw_set_attributes()
        self._window.addnstr(0,pos,string, self._screen_w - pos)

    def draw_footer(self,pos,string):
        self.draw_set_attributes()
        self._window.addnstr(self._screen_h+1,pos,string,self._screen_w - pos)  

    def draw_box(self):
        self.draw_set_attributes()
        self._window.box()    

    def redraw(self,justborder = 0):
        if not justborder:
            self.draw_clear()
            self.redraw_center()
        self.draw_box()
        self.redraw_border()   # "virtual function"

    def draw_line(self,pos,str,cursor=0):
        if pos < 0 or pos >= self._screen_h:
            Warning("out of bounds line drawing\n")

        self.draw_set_attributes(cursor)
           
        self._window.move(pos+1,1)

        str += ' ' * self._screen_w
        str = str[0:self._screen_w]

        
        self._window.addstr(str)        
       
#######################################################
## 
#######################################################

class PanelRoot(Panel):

    global GlobalMusicus
    global GlobalWindowSongPlayer

    def __init__(self, parent, name, x, y , w, h):
        Panel.__init__(self, parent, name, x, y, w, h)
        #self._keymap.bind(12, self.update, ()) # C-l
        self._keymap.bind(ord('q'), sys.exit, (0,), "f10,q:\tQuit",0)
        self._keymap.bind(curses.KEY_F10, sys.exit, (0,), None,0)
        self._keymap.bind(ord('h'), self.toggle_help, (), "h:\tToggle Help Screen",0)
        self._keymap.bind(ord('\t'), self.cycle_next, (), "tab:\tNext Screen",0)
#        self._keymap.bind(curses.KEY_RIGHT, self.cycle_next, (), "c:\tNext Screens",0)

        self._pan.bottom()
        self._pan.hide()
        return
    
    # cycle focus through all children
    
    def cycle_prev(self):
        list = self._children + self._children
        
        for i in range(0,len(list)):

            if list[i]._focus:
                focus = list[i]
                next = list[i+1]
                break
            
        focus.focus_off()
        next.focus_on()
        return
    
    def cycle_next(self):
        list = self._children + self._children
        
        for i in range(len(list)-1,-1,-1):

            if list[i]._focus:
                focus = list[i]
                next = list[i-1]
                break
            
        focus.focus_off()
        next.focus_on()

    def toggle_help(self):
        self.redraw()

        if self._pan.hidden():
            self._pan.show()
            self._pan.top()
        else:
            self._pan.bottom()
            self._pan.hide()
        return
    
    def redraw_border(self):
       self.draw_header(1,"[" +  __version__ + "]")
       self.draw_footer(1,"[Help - Press 'h' to exit]")
       return
   
    def redraw_center(self):
        i=0
        self.draw_line(i,"")
        i+=1
        if 0:
            self.draw_line(i,GlobalLine)
            i += 1
            self.draw_line(i,"Players")
            i += 1
            self.draw_line(i,GlobalLine)
            i += 1

            messages = GlobalPlayers.help_strings()
            for l in messages:
                self.draw_line(i,l)
                i += 1

        messages = self.help_strings()
        for l in messages:
            self.draw_line(i,l)
            i += 1
        while i < self._screen_h:
            self.draw_line(i,"")
            i+=1
        return
    
#######################################################
## Scrollable Panel Class
#######################################################

class ScrollablePanel(Panel):

    FLAG_CURSOR   = 1
    FLAG_MARKED   = 2

    def __init__(self,parent,name, x, y, h, w):
        Panel.__init__(self, parent, name, x, y, h, w)
        self._flags = []
        self._rows = []
        self._cursor_pos = 0
        self._screen_pos = 0
        
        self._keymap.bind(curses.KEY_UP, self.pos_cursor_move, (-1,),"up:\tScroll up",1) 
        self._keymap.bind(curses.KEY_DOWN, self.pos_cursor_move, (1,),"dn:\tScroll down",1) 
        self._keymap.bind(curses.KEY_PPAGE, self.pos_cursor_move, (-h+5,),"pgup:\tScroll page up",1)
        self._keymap.bind(curses.KEY_NPAGE, self.pos_cursor_move, (h-5,),"pgdn:\tScroll page down",1)
        self._keymap.bind(curses.KEY_HOME, self.pos_cursor_move, (-100*1000,),"home:\tScroll home",1)
        self._keymap.bind(curses.KEY_END, self.pos_cursor_move, (100*1000,),"end:\tScroll end",1)
        return

    def pos_is_visible(self,pos):
        if pos < self._screen_pos: return 0
        if pos >= self._screen_pos + self._screen_h: return 0
        return 1

    def pos_make_visible(self, pos):
        assert( pos >= 0 )
        
        if pos < self._screen_pos:
            self._screen_pos = pos
            self.lines_redraw()

        if pos >= self._screen_pos+self._screen_h:
            self._screen_pos = pos - self._screen_h +1
            if self._screen_pos < 0:
                self._screen_pos = 0
            self.lines_redraw()
        return
                
    def pos_cursor_set(self, newpos):
        if len(self._rows) <= 0: return

        assert( self._flags[self._cursor_pos] & PanelList.FLAG_CURSOR )

        self._flags[self._cursor_pos] &= ~PanelList.FLAG_CURSOR
        self.line_redraw( self._cursor_pos )

        if newpos < 0: newpos = 0
        if newpos >= len(self._rows): newpos = len(self._rows) - 1
        # update cursor pos
        self._cursor_pos = newpos

        assert( self._cursor_pos >= 0 )

        self.pos_make_visible( self._cursor_pos )

        self._flags[self._cursor_pos] |= PanelList.FLAG_CURSOR
        self.line_redraw( self._cursor_pos )
        return

    def pos_cursor_move(self,delta):
        self.pos_cursor_set(self._cursor_pos + delta )
        return

    def pos_cursor_get(self):
        return self._cursor_pos

    def pos_random_get(self):
        return whrandom.randint(0, len(self._rows) - 1)
    
    def swap_current(self,delta):
        oldpos = self._cursor_pos
        newpos = oldpos + delta
        maxpos = len(self._rows) - 1
        if oldpos < 0: return
        if newpos < 0: return
        if oldpos > maxpos: return
        if newpos > maxpos: return

        obj = self._lines[oldpos]
        self._lines[oldpos] = self._lines[newpos]
        self._lines[newpos] = obj

        self.redraw_line(oldpos)
        self.pos_cursor_set(newpos)
        
    def line_redraw(self, pos):
        if not self.pos_is_visible(pos): return
        
        if pos < len(self._rows):
            cursor = self._flags[pos] & PanelList.FLAG_CURSOR
            str = self._rows[pos]
        else:
            cursor = 0
            str = ""

        self.draw_line(pos-self._screen_pos ,str,cursor)
        return
                    
    def lines_redraw(self):
        for pos in range(self._screen_h):
            self.line_redraw(self._screen_pos + pos)
        return
    
    def lines_clear(self,redraw=0):
        self._rows = []
        self._flags = []
        self._cursor_pos = 0
        self._screen_pos = 0
        if redraw:
            self.lines_redraw()
        return

    def line_delele(self, pos, redraw=0):
        if self._cursor_pos == pos:
            self.pos_cursor_move(1)

        if self._cursor_pos == pos:
            self.pos_cursor_move(-1)
            
        del self._flags[pos]
        del self._rows[pos]
        
        if self._cursor_pos >= pos:
            self._cursor_pos -= 1
        if redraw:
            self.lines_redraw()
        return

    def line_insert(self, pos, str, redraw=0):

        old_num_rows = len(self._rows)
        self._flags.insert(pos, 0)
        self._rows.insert(pos, str)

        if self._cursor_pos >= pos:
            self._cursor_pos += 1
        if old_num_rows <= 0:
            self._cursor_pos = 0
            self._screen_pos = 0
            self._flags[0] |= PanelList.FLAG_CURSOR
        if redraw:
        #
        # this is an important speed optimization:
        # for complete directory changes
        #
            if pos == old_num_rows and pos != self._cursor_pos: 
                self.line_redraw(pos)
            else:
                self.lines_redraw()
        return

    def line_append(self,str,redraw=0):
        pos = len(self._rows)
        self.line_insert( pos ,str,redraw)
        return pos
    
    def line_string_set(self,pos, str, redraw = 0):
        self._rows[pos] = str
        if redraw:
            self.line_redraw(pos)
        return

    def line_string_get(self,pos,):
        return self._rows[pos] 
                                          
    def redraw_center(self):
        self.lines_redraw()
        return

#######################################################
## 
#######################################################

class PanelList(ScrollablePanel):
    
    def __init__(self, parent, name, x, y, h, w):
        ScrollablePanel.__init__(self, parent, name, x, y, h, w)
        self._playlist = song.Playlist()  
        self._viewmode = song.Viewmode()
   
        self._keymap.bind(ord('v'), self.viev_mode_cycle, (),"v:\tCycle view mode",1)
        return

    def item_string_get(self,n):
        item = self._playlist.item(n)
        return item.fancy_string_get(self._viewmode)
            
    def viev_mode_cycle(self):
        self._viewmode.toggle()

        for i in range(self._playlist.length()):
            self.line_string_set(i,self.item_string_get(i))
        self.redraw()
        return
    
    def items_clear(self):
        self.lines_clear()
        self._playlist.clear()
        return

    def item_append(self,item, redraw):
        str = item.fancy_string_get(self._viewmode)
        self._playlist.append(item)
        return self.line_append(str,redraw)

    def item_delete(self, pos, redraw):
        self.line_delele(pos, redraw)
        self._playlist.delete(pos)
        return

    def item_swap(self, pos1, pos2, redraw):
        self._playlist.swap(pos1,pos2)
        str1 = self.line_string_get(pos1)
        str2 = self.line_string_get(pos2)
        self.line_string_set( pos1, str2, redraw)
        self.line_string_set( pos2, str1, redraw)
        self.pos_cursor_set( pos2)
        return
    
    def item_insert(self,pos,item,redraw):
        str = item.fancy_string_get(self._viewmode)
        self.line_insert(pos,str,redraw)
        self._playlist.insert(pos,item)
        return

    def item_get(self,pos):
        return self._playlist.item( pos )
            
    def item_get_current(self):
        return self._playlist.item( self.pos_cursor_get() )
        
#######################################################
## PLAYBACK WINDOW
#######################################################

class PanelSong(PanelList):

    global GlobalTimer
    global GlobalMusicus
    global ConfigSpecialPlaylistRoot
    
    def __init__(self,parent,name, x, y, h, w):
        PanelList.__init__(self, parent, name, x, y, h, w)
        self._current_item = None
        self._callback_abort = None
        self._callback_finished = None
        self._play_list_root = ConfigSpecialPlaylistRoot
        self._oldpos = 0
        self._paused = 0
        self._skip_bad_songs = 1
        GlobalTimer.add_periodic(0,self.timer_based_polling,())

        self._keymap.bind(ord('-'), self.change_volume, (-3,),"-:\tVol-",0)
        self._keymap.bind(ord('+'), self.change_volume, (+3,),"+:\tVol+",0)
        self._keymap.bind(ord('p'), self.toggle_pause,(),"p:\tToggle pause",0)
        self._keymap.bind(ord('k'), self.kill_song,(),"k:\tStop Song/Playlist",0)
        self._keymap.bind(ord('i'), self.show_info,(),"i:\tshow info for current song",0)
        self._keymap.bind(ord(' '), self.skip_song,(),"space:\tSkip Song",0)
        self._keymap.bind(ord('g'), self.add_to_good_list,(),"g:\tAdd to Goodlist",0)
        self._keymap.bind(ord('b'), self.add_to_bad_list,(),"b:\tAdd to Badlist",0)
        self._keymap.bind(ord('s'), self.toggle_bad_song_skipping,(),"s:\tToggle Bad Song Skipping",0)

        
        self._keymap.bind(ord('l'), self.song_move, (-1000,), "<-,l:\tFast Rewind",0)
        self._keymap.bind(curses.KEY_LEFT, self.song_move, (-1000,), None,0)

        self._keymap.bind(ord('r'), self.song_move, (-1000,), "->,r:\tFast Forward",0)
        self._keymap.bind(curses.KEY_RIGHT, self.song_move, (1000,), None, 0)

        self._bad_list = song.Playlist()
        self._bad_list.load(ConfigSpecialPlaylistRoot + "/bad.m3u")

        self._good_list = song.Playlist()
        self._good_list.load(ConfigSpecialPlaylistRoot + "/good.m3u")

        return

    def toggle_bad_song_skipping(self):
        self._skip_bad_songs = not self._skip_bad_songs
        self.redraw()
        return
    
    def add_to_bad_list(self):
        if self._current_item == None: return
        if self._bad_list.contains( self._current_item ): return
        self._bad_list.append( self._current_item )
        self._bad_list.save()
        self.skip_song()
        return

    def add_to_good_list(self):
        if self._current_item == None: return
        if self._good_list.contains( self._current_item ): return
        self._good_list.append( self._current_item )
        self._good_list.save()
        return

    def show_info(self):
        if self._current_item == None: return
        cmd = ConfigShowInfoScript + " " + self._current_item.description_get() + "&"
        Info("info command: " + cmd + "\n")
        os.system( cmd )
        return

    def kill_song(self):
        if self._current_item:
            self._current_item = None
            Info("killing current song\n")
            GlobalMusicus.stop()
            if self._callback_abort:
                self._callback_abort()
        self.redraw()
        return
    
    def skip_song(self):
        if self._current_item:
            self._current_item = None
            Info("Skipping current song\n")
            GlobalMusicus.stop()
            if self._callback_finished:
                self._callback_finished()
        self.redraw()
        return
    
    def timer_based_polling(self):

        if self._current_item == None: return
        #
        # there appears to be a bug the plugins which
        # may result in negative positions to be returned
        # we fix this by getting a second opinion after some delay 
        #
        pos = self.get_pos()
        if pos < 0:
            self.skip_song()
        if pos != self._oldpos:
            self._oldpos = pos
            self.redraw(1)
        return
    
    def get_pos(self):
        result = GlobalMusicus.getpos()
        if result[0] == 0:
            pos = int(result[1])
            #Info("got pos %d\n" %pos)
            return pos
        else:
            return -1

    def get_length(self):
        result = GlobalMusicus.getlength()
        if result[0] == 0:
            return int(result[1])
        else:
            return -1

    def song_move(self,msec):
        pos = self.get_pos()
        if pos < 0: return
        pos = pos + msec
        if pos < 1: pos = 1
        GlobalMusicus.setpos(pos)
        return
    
    def play_song(self,item,callback_abort,callback_finished):
        self._paused = 0
        if not item.is_song(): return 0

        file =  item.string_get()
        if not os.access(file,os.R_OK): return 0
        
        if self._skip_bad_songs and self._bad_list.contains(item):
            Info("never list match, skipping" + file + "\n")
            return 0

        # if we are already playing a song kill it

        self.kill_song()
        Info("playing " + file + "\n")
        result = GlobalMusicus.play(file)
        if result[0]:
            Warning("player not found for " + file + "\n")
            return 0

        #
        # this delay loop is necessary because???
        # The mpc plug in appears to be buggy 
        #
        
        time.sleep(0.05)

        self.get_length()
        
        while self.get_pos() < 0:
            pass
        
        self._callback_abort = callback_abort
        self._callback_finished = callback_finished
        pos = self.item_append(item,0)
        self.pos_cursor_set(pos) # move to end of list
        
        #self.redraw()
        # !!! this one must be delayed as long as possible since
        # it is the semaphore for the timer based polling
        self._current_item = item
                
        return 1

    def get_volume(self):
        result = GlobalMusicus.getvol()
        if result[0] == 0:
            assert( int(result[1]) <=100)
            assert( int(result[2]) <=100)
            return (int(result[1]) + int(result[2]))/2
        else:
            return 0

    def change_volume(self,delta):
        vol = self.get_volume() + delta
        result = GlobalMusicus.setvol(vol,vol)
        Info("changing volume to %d\n" % vol )
        self.redraw(1)
        return
    
    def toggle_pause(self):
        if self._current_item:
            if self._paused:
                self._paused = 0
                GlobalMusicus.unpause()
            else:
                self._paused = 1
                GlobalMusicus.pause()
            self.redraw_border()    
        return

    def redraw_border(self):
        self.draw_footer(1,"[Song]")
        self.draw_footer(8,"[" + self._viewmode.string() + "]")

        self.draw_footer(20,"[BL %3d]" % self._bad_list.length())
        self.draw_footer(30,"[GL %3d]" % self._good_list.length())

        self.draw_footer(40,"[Vol %3d]" % self.get_volume())

        if self._skip_bad_songs:
            self.draw_footer(50,"[skip]")

        if self._current_item and self._paused:
            self.draw_footer(60,"[paused]")

        if self._current_item:
            self.draw_footer(70,"[sec %4d:%4d]" % ( self.get_pos()/1000,self.get_length()/1000))
        

            #self.draw_footer(10,"%d" % self._song._pid)
            #self.draw_footer(20,"(%d,%d)" %q self._song._result)
        return
    
        
#######################################################
## PLAYLIST EDITOR WINDOW
#######################################################


class PanelListPlaylistEditor(PanelList):

    global GlobalWindowSongPlayer
    global ConfigSpecialPlaylistRoot
    
    def __init__(self,parent,name, x, y, h, w):
        PanelList.__init__(self, parent, name, x, y, h, w)
        self._keymap.bind(ord('\n'), self.play_file, (),"enter:\tPlay file",1)
        #self._keymap.bind(curses.KEY_ENTER, self.play_file, (),"Play file",1)

        self._keymap.bind(ord('t'), self.toggle_track_mode, (),"t:\tToggle Track Mode",1)
        self._keymap.bind(ord('c'), self.new_play_list, (),"c:\tNew Playlist",1)
        
        self._keymap.bind(ord('u'), self.song_move_current, (-1,),"u:\tmove line up",1) 
        self._keymap.bind(ord('d'), self.song_move_current, (1,),"d:\tmove line down",1)
        
        self._keymap.bind(curses.KEY_F6, self.song_insert_current_comment, (),"f6:\tInsert Comment",1)
        self._keymap.bind(ord('x'), self.song_delete_current, (),"x,f8:\tDelete Line",1)
        self._keymap.bind(curses.KEY_F8 , self.song_delete_current, (),None,1)

        self._keymap.bind(curses.KEY_F2, self.save_current_filename, (),"f2:\tSave Playlist",1)
        #self._keymap.bind(curses.KEY_F3, self.save_current_filename, (),"f2:\tSave Playlist",1)
        #self._keymap.bind(curses.KEY_F4, self.del_line_current, (),"f3:\tEdit Line",1)

        self._keymap.bind(ord('m'), self.toggle_playback_mode, (),"r:\tToggle Playback Mode",0)
        
        self._keymap.bind(ord('0'), self.hotkey, (0,),"0,1,2,3,4,5,6,7,8,9:\tplaylist hotkey",0)
        self._keymap.bind(ord('1'), self.hotkey, (1,),None,0)
        self._keymap.bind(ord('2'), self.hotkey, (2,),None,0)
        self._keymap.bind(ord('3'), self.hotkey, (3,),None,0)
        self._keymap.bind(ord('4'), self.hotkey, (4,),None,0)
        self._keymap.bind(ord('5'), self.hotkey, (5,),None,0)
        self._keymap.bind(ord('6'), self.hotkey, (6,),None,0)
        self._keymap.bind(ord('7'), self.hotkey, (7,),None,0)
        self._keymap.bind(ord('8'), self.hotkey, (8,),None,0)
        self._keymap.bind(ord('9'), self.hotkey, (9,),None,0)

        self._mode_playback = ["normal","shuffle","repeat"] 
        self._current_play_pos = -1
        self._mode_track = 1
        return

    def hotkey(self,n):
        self.load_file( ConfigSpecialPlaylistRoot + ("/%d.m3u" % n) )
        self.play_file()
        return
    
    def toggle_playback_mode(self):
        self._mode_playback = self._mode_playback[1:] + self._mode_playback[0:1]
        self.redraw()
        return

    def toggle_track_mode(self):
        if self._mode_track == 0:
            self._mode_track = 1
        else:
            self._mode_track = 0
        self.redraw()
        return
    
    def get_next_song(self):
        if self._playlist.length() <= 0: return None
        while 1:
            if  self._mode_playback[0] == "shuffle":
                self._current_play_pos = self.pos_random_get()
            else:
                self._current_play_pos += 1
                if self._current_play_pos >= self._playlist.length():
                    if self._mode_playback == "repeat":
                        self._current_play_pos = 0
                    else:
                        return None                

            obj = self.item_get(self._current_play_pos)
            if  obj.is_song():
                return obj

    def play_finish(self):
        for i in range(100):
            obj = self.get_next_song()
            if obj == None:
                break
            if GlobalWindowSongPlayer.play_song(obj,self.play_abort,self.play_finish):
                if self._mode_track:
                    self.pos_cursor_set(self._current_play_pos)
                break
        return
    
    def play_abort(self):
        return
    
    def play_file(self):
        self._current_play_pos = self._cursor_pos - 1 
        self.play_finish()
        return
    
    def load_file(self,filename):
        if self._playlist.dirty():
            Warning("overwrting dirty playlist\n")

        self.items_clear()
        self._playlist.load(filename)
        for item in  self._playlist.items():
            str = item.fancy_string_get(self._viewmode)
            self.line_append(str)
        self.redraw()
        return
    
    def new_play_list(self):
        if self._playlist.dirty():
            Warning("overwrting dirty playlist\n")

        self.items_clear()        
        self.redraw()
        return
    
    def save_file(self,filename):
        self._playlist.save(filename)
        self.redraw_border()
        return
    
    def save_current_filename(self):
        self._playlist.save()
        self.redraw_border()
        return
    
    def song_insert_current_comment(self):
        pos = self.pos_cursor_get()
        item = song.Item("#")
        self.item_insert(pos,item,1)
        self.redraw_border()
        return
    
    def song_delete_current(self):

        if self._playlist.length() > 0:
            pos = self.pos_cursor_get()
            self.item_delete(pos,1)
            self.redraw() # FIXME there is a bug somewhere, so we use the big hammer for now
            #self.redraw_border()
        return

    def song_move_current(self,delta):
        pos1 = self.pos_cursor_get()
        pos2 = pos1 + delta
        if pos2 < 0: return
        if pos2 >= self._playlist.length(): return
       
        self.item_swap(pos1,pos2,1)
        self.redraw_border()
        return

    def song_append(self,item):
        self.item_append(item,1)
        self.redraw_border()
        return

    def redraw_border(self):
        self.draw_header(1,"[" + self._playlist.filename() +"]")

        self.draw_footer(1,"[List]")
        self.draw_footer(8,"[" + self._viewmode.string() + "]")

        self.draw_footer(19,"[" + self._mode_playback[0] + "]")

        if self._mode_track:
            self.draw_footer(30,"[track]")
                
        self.draw_footer(40,"[" + ("%d" % self._playlist.length()) + "]")
        marker = ' '
        if self._playlist.dirty(): marker = '*'
        self.draw_footer(50,"[" + marker + "]")
        return

#######################################################
## FILE BROWSER WINDOW
#######################################################

def entry_compare(e1,e2):
    t1 = e1.type_get()
    t2 = e2.type_get()
    if t1 > t2: return 1
    elif t1 < t2: return -1
    
    p1 = e1.string_get()
    p2 = e2.string_get()
    
    if p1 > p2: return 1
    elif p1 < p2: return -1
    else: return 0
    
        
class PanelListFilebrowser(PanelList):

    global GlobalWindowSongPlayer
    global GlobalWindowPlaylistEditor

    def __init__(self,parent,name, x, y, h, w, directory):
        PanelList.__init__(self, parent, name, x, y, h, w)

        #self._keymap.bind(curses.KEY_ENTER, self.enter_button, (),"Change directory/Play song",1)
        self._keymap.bind(ord('\n'), self.enter_button, (),"enter:\tChange directory/Play song",1)

        self._keymap.bind(curses.KEY_BACKSPACE, self.change_dir_up, (),"bs,u:\tChange directory up",1)
        self._keymap.bind(ord('u'), self.change_dir_up, (),None,1)


        self._keymap.bind(curses.KEY_F4, self.edit_playlist, (),"f4:\tEdit playlist",1)
        self._keymap.bind(curses.KEY_F8, self.delete_file, (),"f8:\tDelete file",1)

        self._keymap.bind(ord('t'), self.toggle_playable_only, (),"t:\tToggle List Playable ",1)

        self._keymap.bind(curses.KEY_F5, self.append_to_playlist, (0,),"f5:\tAppend to playlist",1)
        self._keymap.bind(curses.KEY_F6, self.append_to_playlist, (1,),"f6:\tAppend to playlist recursively",1)

        self._playable_only = 1
        if not os.path.isdir(directory):
            Warning("cannot find start directory %s\n" % directory)
            directory = "/"
        
        self.change_directory(directory)

    def change_directory(self, dir, prefered="$$"):

        Info("change dir to " + dir + " prefered: " + prefered + "\n")
        
        if not os.access(dir,os.R_OK | os.X_OK): return

        dir = os.path.normpath(dir)
        self.items_clear()
        self._cwd = dir
        itemlist = []

        #item = Item( os.path.join(dir,os.pardir) )
        #itemlist.append( item )

        for entry in os.listdir(dir):
            pathname = os.path.join(dir,entry)
            item = song.Item(pathname)
            if self._playable_only and item.type_get() == song.Item.TYPE_FILE:
                continue
            itemlist.append( item )

        itemlist.sort( entry_compare )

        prefered_pos = None
        for entry in itemlist:
            pos = self.item_append(entry,0)
            if prefered == entry.string_short_get():
                prefered_pos = pos 

        if prefered_pos:
            self.pos_cursor_set(prefered_pos)
            
        self.redraw()
        return
    
    def change_dir_up(self):
        dummy, currdir = os.path.split(self._cwd)
        #assert (not currdir)
        newdir =  os.path.join(self._cwd,os.pardir)
        if not os.path.isdir(newdir): return
        self.change_directory(newdir,currdir)
        return
    
    def change_dir(self):
        obj =  self.item_get_current()
        if not obj.is_dir(): return
        newdir =  obj.string_get()
        self.change_directory(newdir)
        return

    def  enter_button(self):
        obj =  self.item_get_current()
        if obj.is_song():
            GlobalWindowSongPlayer.play_song(obj, None, None)
        elif obj.is_dir():
            self.change_dir()
        return
                
    def toggle_playable_only(self):
        if self._playable_only:
            self._playable_only = 0
        else:
            self._playable_only = 1
        self.change_directory( self._cwd )
        return

    def delete_file(self):
        file =  self.item_get_current().string_get()
        if os.path.isdir(file): return
        assert(0)
        

    def add_to_playlist_generic(self, filename):
        obj = song.Item(filename)
        if obj.is_song():
            GlobalWindowPlaylistEditor.song_append(obj)
            
        elif obj.is_dir():
            
            try:
                for file in os.listdir(filename):
                    pathname = os.path.join(filename, file)
                    self.add_to_playlist_generic(pathname)
            except IOError:
                pass

#        elif obj.type_get() == Item.TYPE_LIST:
#            try:
#                file = open(filename)
#                for filename in map(string.strip, file.readlines()):
#                    if patternComment.match(filename): continue
#                    type = FileType(filename)
#                    entry = PlaylistEntry(type, filename)
#                    self.append(entry)
#                file.close()
#            except IOError:
#                app.status(_("IOError while reading playlist " + directory), 1)


    def append_to_playlist(self,recursive):
        obj =  self.item_get_current()
        if recursive or obj.is_song():
            self.add_to_playlist_generic(obj.string_get())
        self.pos_cursor_move(1)
        GlobalWindowPlaylistEditor.redraw()
        return
    
    def edit_playlist(self):
        obj =  self.item_get_current()
        if not obj.is_playlist() : return
        file =  self.item_get_current().string_get()
        GlobalWindowPlaylistEditor.load_file(file)
        return
    
    def redraw_border(self):
        self.draw_header(1,"[" + self._cwd +"]")
        self.draw_footer(1,"[Files]")
        self.draw_footer(10,"[" + self._viewmode.string() + "]")

        if self._playable_only:
            self.draw_footer(20,"[playable]")

        self.draw_footer(35,"[" + ("%d" % self._playlist.length()) + "]")
        return
#######################################################
## 
#######################################################

def usage():
    usage = _("Usage: %s [ filename | playlist.m3u ] ...\n\n\tpress h for a help screen\n")
    sys.stderr.write(usage % sys.argv[0])
    return

#######################################################
## 
#######################################################

def cursesmain(stdscr):

    global GlobalTimer
    global GlobalMusicus
    global GlobalWindowRoot
    global GlobalWindowSongPlayer
    global GlobalWindowPlaylistEditor
    global GlobalWindowFileBrowser

    global ConfigMusicRoot

    Info("##################################################\n")
    Info("starting cjukebox\n")
    Info("##################################################\n")
    
    # make getch nonblocking
    stdscr.nodelay(1)

    # make cursor invisible
    # causes problems on some systems
    curses.curs_set(0)

    if curses.has_colors():
            
        curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLUE)
        Window.ColorFocus.append(curses.color_pair(1))
        curses.init_pair(2, curses.COLOR_BLUE, curses.COLOR_WHITE)
        Window. ColorFocus.append(curses.color_pair(2))
        curses.init_pair(3, curses.COLOR_BLACK, curses.COLOR_RED)
        Window.ColorFocus.append(curses.color_pair(3))
        
        curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLUE)
        Window.ColorNoFocus.append(curses.color_pair(1))
        curses.init_pair(2, curses.COLOR_BLUE, curses.COLOR_WHITE)
        Window. ColorNoFocus.append(curses.color_pair(2))
        curses.init_pair(3, curses.COLOR_BLACK, curses.COLOR_RED)
        Window.ColorNoFocus.append(curses.color_pair(3))
    
    #
    # Prepare Timer
    #

    GlobalTimer = TimerEvents()

    #
    # Initialize Musicus
    #
    
    GlobalMusicus = musicus.Musicus()

    #
    #
    #
    
    GlobalWindowRoot = PanelRoot(None,"Root Window",0,0,curses.COLS,curses.LINES)
    GlobalWindowRoot.focus_on()

    #

    GlobalWindowPlaylistEditor = PanelListPlaylistEditor(GlobalWindowRoot,"Playlist Editor",
                curses.COLS/2,0,curses.COLS - curses.COLS/2,curses.LINES - 8)
    GlobalWindowPlaylistEditor.focus_off()

    #

    GlobalWindowFileBrowser = PanelListFilebrowser(GlobalWindowRoot,"File Browser",
                                                   0,0,curses.COLS/2,curses.LINES - 8,ConfigMusicRoot)
    GlobalWindowFileBrowser.focus_on()

    #
    
    GlobalWindowSongPlayer = PanelSong(GlobalWindowRoot,"Current Song Window",
                                       0,curses.LINES -8 ,curses.COLS,8)
    GlobalWindowSongPlayer.focus_off()

    #xx = ModalDialog(10,10,curses.COLS -10,20)
    
    ######################
    ## MAIN LOOP
    #####################
    
    try:
        while 1:
            # for performance reasons
            time.sleep(0.01)
            #GlobalWindowRoot.update()
            panel.update_panels()
            curses.doupdate()
            GlobalTimer.poll()
            c = stdscr.getch()
            if c >= 0:
                GlobalWindowRoot.process_key(c)
    finally:
         GlobalWindowSongPlayer.kill_song()
         GlobalMusicus.kill()

    return

#######################################################
## 
#######################################################

if __name__ == "__main__": 
    try:
        opts, args = getopt.getopt(sys.argv[1:], "")
    except:
        usage()
        sys.exit(1)

    whrandom.seed()

    if 1:
        curses.wrapper( cursesmain )
    else:
        import profile
        import pstats

        profile.run("curses.wrapper( cursesmain )","run.prof")    
        p = pstats.Stats("run.prof")
        p.sort_stats('cumulative').print_stats(20)
        p.sort_stats('time').print_stats(20)

#######################################################
## 
#######################################################

