#######################################################
## 
#######################################################
import os
import re

import cddb

#######################################################
## 
#######################################################

ConfigMagicFiles = ["id","index","flags","toc","toc.1","toc.2","toc.3"]

#######################################################
## 
#######################################################

class Viewmode:

    modes =[ ("cddb",0,1,1), ("cddbfull",1,1,1),("short",0,0,1),("full",1,0,0) ]
             
    def __init__(self):
        self._mode = 0

    def string(self):
        return Viewmode.modes[ self._mode][0]

    def bits(self):
        bits = Viewmode.modes[ self._mode][1:]
        assert( len(bits) == 3)
        return bits
    
    def toggle(self):
        self._mode += 1
        if self._mode == len(Viewmode.modes):
            self._mode = 0

#######################################################
## Item Object for use with PanelList Object, etc
#######################################################

class Item:

    TocContainer = cddb.TocContainer()


    use_toc = 1
    
    PatternEmptyLine = re.compile(r"^[ \t\r]*$")
    PatternCommentLine = re.compile(r"^#.*$")
    PatternPlayList = re.compile(r"^.*\.m3u$", re.I)


    TYPE_INVALID = 0
    TYPE_DIR  = 1
    TYPE_LIST = 2
    TYPE_SONG = 3

    TYPE_FILE = 4
    TYPE_COMMENT = 5
    TYPE_BADFILE = 6
    TYPE_EMPTY = 7
        
    def __init__(self, string):
        self._cool = 1
        self._toc = None
        self._string = string
        self._size = 0
        
        match = Item.PatternEmptyLine.search(string)
        if match:
            self._type = Item.TYPE_EMPTY
            return

        match = Item.PatternCommentLine.search(string)
        if match:
            self._type = Item.TYPE_COMMENT
            return

        match = Item.PatternPlayList.search(string)
        if match:
            self._type = Item.TYPE_LIST
            return

        if os.path.isdir(string):
            self._type = Item.TYPE_DIR
            if Item.use_toc:
                self._toc = Item.TocContainer.find_toc(string)
            return

        if os.path.exists(string):
            self._size = os.path.getsize(string)
            (dirname,filename) = os.path.split(string)
            if filename in ConfigMagicFiles:
                self._type = Item.TYPE_FILE
            else:
                self._type = Item.TYPE_SONG            
#            player = GlobalPlayers.find_player(string)
#            if player == None:
#                self._type = Item.TYPE_FILE

            self._toc = Item.TocContainer.find_toc(dirname)
            return 

        else:
            self._type = Item.TYPE_BADFILE
        return
    
    def string_get(self):
        return self._string

    def string_short_get(self):
        dummy, string = os.path.split(self._string)
        return string

    def decorate_string(self,string):
        if self._type == Item.TYPE_EMPTY:
            return "  " + string

        elif self._type == Item.TYPE_SONG:
            if self._toc == None:
                return "s " + string  
            else:
                return "S " + string 
            
        elif self._type == Item.TYPE_FILE:
            return "f " + string

        elif self._type == Item.TYPE_COMMENT:
            return "c " + string
        
        elif self._type == Item.TYPE_LIST:
            return "p " + string

        elif self._type == Item.TYPE_BADFILE:
            return "b " + string 

        elif self._type == Item.TYPE_DIR:
            if self._toc == None:
                return "d " + string + os.sep 
            else:
                return "D " + string + os.sep 

        assert(0)

    def decorate_string2(self,string):
        if self._type == Item.TYPE_EMPTY:
            return string

        elif self._type == Item.TYPE_SONG:
            if self._toc == None:
                return string + "*"  
            else:
                return string + "*"
            
        elif self._type == Item.TYPE_FILE:
            return string

        elif self._type == Item.TYPE_COMMENT:
            return  string
        
        elif self._type == Item.TYPE_LIST:
            return string + "@"

        elif self._type == Item.TYPE_BADFILE:
            return string + "!" 

        elif self._type == Item.TYPE_DIR:
            if self._toc == None:
                return string + os.sep 
            else:
                return string + os.sep 

        assert(0)
    

    def fancy_string_get(self,viewmode):
        
        long, cddb, decoration = viewmode.bits()
       
        string = self.string_get()

        if not long:
            if self._type == Item.TYPE_SONG or self._type == Item.TYPE_LIST or self._type == Item.TYPE_DIR:
                string = self.string_short_get()

        if cddb and self._toc:
            if self._type == Item.TYPE_DIR:
                string = "[" + self._toc.title_get() + "]" 
            if self._type == Item.TYPE_SONG:
                (track,ext,time) = self._toc.track_get(self._string)
                if track:
                    string = "[" + ("%2d:%02d " % divmod(time,60)) + track + "]"
                    if long:
                        string = "[" + self._toc.title_get() + "]" + string 

        if decoration:
            string = self.decorate_string(string)
        return string

    def description_get(self):
        string = self.string_get()
        if self._toc:
            (track,ext,time) = self._toc.track_get(self._string)
            artist = self._toc.artist_get()
            string = track + ". by " + artist
        else:
            string = self.string_short_get()

        string = re.sub(r'[_]', " ", string)            
        string = re.sub(r'[/:()]', ".", string)            
        string = re.sub(r'\'', r'\'', string)
        string = re.sub(r'\&', r'and', string)
        return string
            
    def type_get(self):
        return self._type

    def is_song(self):
        return self._type == Item.TYPE_SONG

    def is_dir(self):
        return self._type == Item.TYPE_DIR

    def is_playlist(self):
        return self._type == Item.TYPE_LIST
    
#######################################################
## PLAYLIST
#######################################################

class Playlist:
    
    def __init__(self):
        self._items = []
        self._filename = "noname"
        self._dirty = 0
        return

    def clear(self):
        self._items = []
        self._filename = "noname"
        self._dirty = 0
        return
    
    def contains(self,item):
        for i in self._items:
            if item.string_get() == i.string_get(): return 1
            return 0

    def item(self,n):
        return self._items[n]

    def filename(self):
        return self._filename

    def items(self):
        return self._items

    def append(self,item):
        self._items.append( item )
        self._dirty = 1
        return

    def swap(self,pos1,pos2):
        item1 = self._items[pos1]
        item2 = self._items[pos2]
        self._items[pos1] = item2
        self._items[pos2] = item1
        return
    
    def insert(self,pos,item):
        self._items.insert(pos, item )
        self._dirty = 1
        return

    def delete(self,pos):
        del self._items[pos]
        return
    
    def length(self):
        return len(self._items)

    def dirty(self):
        return self._dirty
        
    def load(self,filename):
        self._items = []
        self._filename = filename
        self._dirty = 0            
        
        try:
            input = open(filename, "r")
            lines = input.readlines()
            input.close()
        except:
            return
        
        for l in lines:
            item = Item(l[:-1])
            self._items.append(item);
        return
   
    def save(self,filename=None):
        #try:
        if not filename:
            filename = self._filename
        input = open(filename, "w+")
        for i in self._items:
            input.write( i.string_get()+"\n"); 
        input.close()
        self._dirty = 0
        return

    

